/*******************************************************************************
*                                 qdos_c                                       *
*                                 ~~~~~~                                       *
*     QDOS/SMS initialization and graphic routines for POV Ray Trace v3.1      *
*                      by Thierry Godefroy 1994-2000.                          *
*                                                                              *
*            Aurora video modes support by Franois Lanciault.                 *
*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <qdos.h>
#include <qptr.h>
#include <fcntl.h>
#include <ctype.h>
#include "frame.h"
#include "povproto.h"
#include "povray.h"
#include "optin.h"

#define INFINITE (timeout_t) -1

/* Definition des mode d'affichage de la carte aurora */

#define C256R512X256    24
#define C256R512X384    152
#define C256R640X320    25
#define C256R640X480    153
#define C4R512X256      0
#define C4R512X384      128
#define C4R640X320      1
#define C4R640X480      129
#define C4R768X384      2
#define C4R768X576      130
#define C4R1024X512     3
#define C4R1024X768     131

/* Variables externes (dclares dans povray_c) */

extern char *Option_String_Ptr;
extern Opts opts;
extern int Display_Started;
extern char *Library_Paths[];
extern int Library_Path_Index;

/* Type for the GD2 stipples */

typedef struct GD2_STIPPLE {
        long    colour1;
        long    colour2;
        long    stipple;
        } GD2_STIPPLE_t;

/* Variables utilises dans qdos_c */

static int redirected;
static int console_in, console_out, console_err;
static chanid_t stdin_id, stdout_id, stderr_id;
static struct QLRECT coord = { 2, 1, 0, 0 };
static struct GD2_STIPPLE stipple = { 0, 0, -1 };
static struct REGS inp_par;
static struct REGS out_par;
static double Kx, Ky;
static short Q40mode;
static short AurMode;
static short graphic_mode = 0;

/* Fonctions dfinies dans q40_asm */

int Q40_test(void);
void Q40_mode(short mode);
void Q40_cls(void);
int Q40_plot(short mode, short x, short y, short Green, short Red, short Blue);

/* Fonctions dfinies dans aurora_asm */

int Aurora_test(void);
void Aurora_mode(short mode);
void Aurora_cls(void);
int Aurora_plot(short mode, short x, short y, short Green, short Red,
                short Blue);

/* Transformation automatique des noms de fichiers UNIX en QDOS */

int (*_Open)(const char *, int, ...) = qopen;

/* Nom du programme, version QDOS, copyright */

char _prog_name[]   = "POV";
char _version[32]   = "v3.1g (" __DATE__ ")";
char _copyright[24] = "QDOS/SMS port";

/* Paramtres de la fentre d'affichage */

void (*_consetup)() = NULL;

struct WINDOWDEF _condetails      = { 208, 1, 0, 7, 512, 256, 0, 0};
static struct QLRECT win_outline  = { 512, 256, 0,  0 };
static struct WM_wsiz max_size    = { 512, 256, 0,  0 };

/* Prompt lors d'un exit() */

char *_endmsg = "\nPress a key to exit.";

/* Besoins et limitations en mmoire */

long _stack   =   65L * 1024L;
long _mneed   =   16L * 1024L;
long _memmax  = 9999L * 1024L;
long _memincr =   16L * 1024L;
long _memqdos =   24L * 1024L;

/* Routine de lecture/changement du mode d'affichage:
    - si m = -1 retourne le mode d'affichage actuel;
    - sinon change le mode et retourne TRUE si russi, FALSE sinon.
*/

static short do_mode(short m)
{
    short a = -1;
    short n =  m;

    mt_dmode(&n, &a);
    if (n == 0)
        n = 4;

    if (m == -1)
        return(n);

    a = -1;
    n = -1;
    mt_dmode(&n, &a);
    if (n == 0)
        n = 4;

    if (n == m) {
        graphic_mode = n;
        return TRUE;
    } else
        return FALSE;
}

/* Routine de dessin de la fentre d'affichage (si non redirige) */

static void draw_stdout(void)
{
    short m;
    char  c;

    if (!redirected) {

        /* Passer en mode 4 ou rester en mode 2 (SMS2) */

        m = do_mode(-1);
        if (m != 4 && m != 2)
            do_mode(4);

        /* Redessiner la fentre de stdout */

        iop_outl(stdout_id, INFINITE, 0, 0, 0, &win_outline);
        sd_wdef(stdout_id, INFINITE, 0, 0, &win_outline);
        consetup_title(stdout_id, &_condetails);
        sd_setin(stdout_id, INFINITE, 4);
        sd_cure (stdin_id, INFINITE);
        io_fbyte(stdin_id, 0, &c);
    }
}

/* Routine d'initialisation du programme */

void QDOS_init(void)
{
    /* Stocker les identificateurs de canaux QDOS de stdout */

    stdout_id = fgetchid(stdout);
    stdin_id  = fgetchid(stdin);
    stderr_id = fgetchid(stderr);

    /* Flags de redirection et de type de priphrique */

    redirected  = isnoclose(fileno(stdout));
    console_out = iscon(stdout_id, INFINITE);
    console_in  = iscon(stdin_id, INFINITE);
    console_err = iscon(stderr_id, INFINITE);

    /* Chercher la taille maximale de l'cran (si l'EP est prsent !) */

    iop_flim(stdout_id, INFINITE, &max_size);

    draw_stdout();
}

/* Routine de sortie du programme */

void QDOS_exit(void)
{
/*  _endmsg = NULL; */
}

/* Routine de test pour sortie immdiate du programme: active si option "+x" */

void QDOS_test_abort(void)
{
    char c1 = 0;
    char c2 = 0;

    if (console_in && (opts.Options & EXITENABLE)) {

        while (io_fbyte(stdin_id, 0, &c1) != -1)
            c2 = c1;

        /* si caractre CTRL Q, quitter */

        if (c2 == '\021') {

            if (Display_Started) {
                if (opts.DisplayFormat == '9') {
                    Q40_cls();
                    Q40_mode(1);
                }

                if (opts.DisplayFormat == '8') {
                    Aurora_cls();

                    /* Remettre l'affichage dans le mode antrieur, */
                    /* on suppose que le nombre de couleurs tait de 4. */
                    /* Pour la rsolution, on regarde la structure max_size */

                    switch (max_size.xsize) {
                        case 512:
                            if (max_size.ysize == 3/4*512)
                                AurMode = C4R512X384;
                            else
                                AurMode = C4R512X256;
                            break;
                        case 640:
                            if (max_size.ysize == 3/4*640)
                                AurMode = C4R640X480;
                            else
                                AurMode = C4R640X320;
                            break;
                        case 768:
                            if (max_size.ysize == 3/4*768)
                                AurMode = C4R768X576;
                            else
                                AurMode = C4R768X384;
                            break;
                        case 1024:
                            if (max_size.ysize == 3/4*1024)
                                AurMode = C4R1024X768;
                            else
                                AurMode = C4R1024X512;
                            break;
                        default:
                            AurMode = C4R512X256;
                    }
                    Aurora_mode(AurMode);
                }
                draw_stdout();
            }

            printf("\n\nAborted !\n");
            exit(-1);
        }
    }
}

/* Routine de test permettant de continuer les calculs sans tenter d'crire
   les messages de l'option "+v" sur une fentre gele, en arrire plan, ou
   pendant un affichage graphique.
*/

static int isnotlocked(void)
{
    if (console_err) {
        if (Display_Started)
            return(FALSE);
        else
            return(sd_cure(stderr_id, 0) == 0);
    }
    else
        return(TRUE);
}

void QDOS_console_print(char *s)
{
    if (isnotlocked()) {
        fprintf(stderr, s);
        fflush(stderr);
    }
}

/* Routine d'initialisation de l'affichage graphique (option "+dN"):
   si N=0, couleurs "vraies" (dominantes rouges/vertes/bleues respectes);
   si N=1, niveaux de gris, chelle linaire;
   si N=2, niveaux de gris, chelle linaire, faible intensit (0  50%);
   si N=3, niveaux de gris, chelle linaire, intensit moyenne (25%  75%);
   si N=4, niveaux de gris, chelle linaire, forte intensit (50%  100%);
   si N=5, niveaux de gris, chelle logarithmique de base 2 (0  100%).
   si N=8, 256 couleurs (Aurora seulement).
   si N=9, 65536 couleurs (Q40 seulement).
*/

void QDOS_display_init(int Width, int Height)
{
    struct QLRECT winsiz;
    int max_x = 1024;
    int max_y = 512;

    if (redirected) {

        /* Il est hors de question de modifier une fentre redirige (le
           job auquel elle appartient pourrait bien s'en offusquer ! */

        printf("\nGraphic displaying on a redirected window is not allowed !\n");
        exit(-1);

    } else {

        /* Vrification de la validit du format demand */

        switch (opts.DisplayFormat) {
            case '0':
            case '1':
            case '2': /* Pour les formats 2  5, la validit est vrifie */
            case '3': /* aprs le passage dans le mode graphique disposant du */
            case '4': /* plus grand nombre de couleurs possibles; ces formats */
            case '5': /* ne sont utilisables qu'avec 16 couleurs ou moins. */
                break;
            case '8':
                if (Aurora_test() == 1)
                    break;
            case '9':
                if (Q40_test() == 1)
                    break;
            default:
                printf("\n\nBad display format !\n");
                exit(-1);
        }

        sd_bordr(stdout_id, INFINITE, 0, 0);

        if (opts.DisplayFormat == '9' || opts.DisplayFormat == '8') {

            /* Ce qui suit est trs malpropre mais les pilotes d'cran n'tant
               pas prts pour la Q40 et Aurora, il faut bien se dbrouiller
               autrement...

               Juin 2000: les pilotes pour Q40 et QXL existent: le mode +d9 est
                          donc obsolte (utiliser +d0  la place); il est
                          toutefois conserv pour ceux qui ne disposent pas de
                          SMSQ/E mais/ou utilisent le QDOS "Classic".
                          Pour Aurora, le mode +d8 deviendra lui aussi bientt
                          obsolte, le mode +d0 devrait galement fonctionner
                          en 256 couleurs pour cette carte ds que le pilote
                          GD2 sera disponible. Le mode +d8 restera utile  ceux
                          n'ont pas SMSQ/E mais/ou utilisent QDOS ou Minerva...

               On commence par vrouiller tout l'cran en mode 4 de faon 
               viter de corrompre l'cran en mode 256 ou 65536 couleurs.
            */

            if (opts.DisplayFormat == '9')
                Q40_mode(1);

            winsiz.q_width  = max_size.xsize;
            winsiz.q_height = max_size.ysize;
            winsiz.q_x      = max_size.xorg;
            winsiz.q_y      = max_size.yorg;

            sd_wdef(stdout_id, INFINITE, 0, 0, &winsiz);
            iop_outl(stdout_id, INFINITE, 0, 0, 0, &max_size);

            sd_clear(stdout_id, INFINITE);

            /* Inhiber l'affichage du curseur */

            sd_curs(stdout_id, INFINITE);

            if (opts.DisplayFormat == '9') {                /* Q40 */

                /* On tient compte de la taille de l'image pour choisir le mode
                   le plus appropri...
                 */

                if (Width > 512 || Height > 256)
                    Q40mode = 3;
                else {
                    Q40mode = 2;
                    max_x = 512;
                    max_y = 256;
                }
                Q40_mode(Q40mode);
                Q40_cls();
            } else {                                        /* Aurora */
                max_x = 512;            /* d'autres rsolutions sont */
                max_y = 480;            /* disponibles, mais on se limite */
                AurMode = C256R640X480; /*  512 x 480 pour l'instant. */
                Aurora_mode(AurMode);   /* il faut demander 640x480 pour */
                                        /* avoir 512x480 ... */
                Aurora_cls();
            }

            /* Coefficients d'chelle par dfaut */

            Kx = 1.0;
            Ky = 1.0;

            if (Width > max_x)
                Kx = ((double) max_x) / ((double) Width);

            if (Height > max_y)
                Ky = ((double) max_y) / ((double) Height);

        } else {

            /* Pas de bidouille dans ce qui suit: on utilise les pilotes
             * "lgaux" de QDOS/Minerva/SMS2/SMSQ, ARGOS ou SMSQ/E...
             */

            /* Mettre l'ordinateur dans le mode possdant le plus de couleurs
               possibles (de 16 millions  4 selon la machine et le type de
               pilote d'cran disponible). En cas d'chec (ex: SMS2 en noir et
               blanc), on restera simplement dans le mode courant (par exemple:
               MODE 2 pour SMS2).
            */
            if (!do_mode(64))                       /* 16 millions */
                if (!do_mode(33))                   /* 65536 au format Q40 */
                    if (!do_mode(32))               /* 65536 au format QXL */
                        if (!do_mode(16))           /* 256 */
                            if (!do_mode(12))       /* 16 (Thor XVI) */
                                if (!do_mode(8))    /* 8 couleurs */
                                    do_mode(4);     /* 4 couleurs */

            if (graphic_mode > 12 && opts.DisplayFormat > '1') {
                printf("\n\nBad display format !\n");
                exit(-1);
            }

            /* Coefficients d'chelle par dfaut */

            Kx = 1.0;
            Ky = 1.0;

            /* Dimensionner la fentre d'affichage */

            winsiz.q_width  = Width;
            winsiz.q_height = Height;
            winsiz.q_x      = 0;
            winsiz.q_y      = 0;

            if (sd_wdef(stdout_id, INFINITE, 0, 0, &winsiz) != 0) {

                winsiz.q_width  = max_size.xsize;
                winsiz.q_height = max_size.ysize;
                winsiz.q_x      = max_size.xorg;
                winsiz.q_y      = max_size.yorg;

                sd_wdef(stdout_id, INFINITE, 0, 0, &winsiz);
                iop_outl(stdout_id, INFINITE, 0, 0, 0, &max_size);

                if (Width > max_size.xsize)
                    Kx = ((double) max_size.xsize) / ((double) Width);

                if (Height > max_size.ysize)
                    Ky = ((double) max_size.ysize) / ((double) Height);
            } else {

                iop_outl(stdout_id, INFINITE, 0, 0, 0, &winsiz);
            }

            sd_clear(stdout_id, INFINITE);

            /* Inhiber l'affichage du curseur */

            sd_curs(stdout_id, INFINITE);

            /* Initialiser certains paramtres pour display_plot() */

            if (graphic_mode <= 12) {
                /* Pour un mode 16 couleurs ou moins, on utilise l'appel
                   systme classique de trac bloc. */
                inp_par.D0 = 0x2E;
            } else {
                /* Pour les modes 256 couleurs et plus, on utilise l'appel
                   du pilote GD2 de trac de bloc. */
                inp_par.D0 = 0x5D;
                inp_par.A2 = (addreg_t) &stipple;   /* Adresse du pattern. */
                coord.q_width = 1;                  /* Bloc large d'un point */
            }

            inp_par.D3 = -1L;                   /* Temporisation */
            inp_par.A0 = (addreg_t) stdout_id;  /* Id de canal de la fentre */
            inp_par.A1 = (addreg_t) &coord;     /* Adresse des coordonnes */
        }
    }
}

/* Routine d'affichage d'un point */

void QDOS_display_plot(int x, int y, unsigned char Red, unsigned char Green,
                       unsigned char Blue, unsigned char Alpha)
{
    int color, intensity, R, G, B, x0, y0;

    R = (int) Red;
    G = (int) Green;
    B = (int) Blue;

    intensity = (B + R + G) / 3;

    x0 = (int) (Kx * ((double) x));
    y0 = (int) (Ky * ((double) y));

    switch (opts.DisplayFormat) {

        case '0':   /* affichage en couleurs "vraies" */
            if (graphic_mode <= 12) {
                if (intensity == 0)
                    color = 0;
                else {
                    color  = ((B / intensity) >= 1);
                    color += ((R / intensity) >= 1) * 2;
                    color += ((G / intensity) >= 1) * 4;
                }
                color += (intensity < 128) * 32768;     /* pour MODE 12 */
            } else
                    color  = (R * 65536 + G * 256 + B) << 8;
            break;
        case '1':   /* niveaux de gris, chelle linaire (0  100%) */
            if (graphic_mode <= 12) {
                color = intensity / 16;
                color = color / 2 + 32768 * (1 - color % 2);
            } else
                color = (intensity * (65536 + 256 + 1)) << 8;
            break;
        case '2':   /* niveaux de gris, chelle linaire (0  50%) */
            color = intensity / 8;
            if (color > 15)
                color = 15;
            color = color / 2 + 32768 * (1 - color % 2);
            break;

        case '3':   /* niveaux de gris, chelle linaire (25  75%) */
            color = (intensity - 64) / 8;
            if (color < 0)
                color = 0;
            if (color > 15)
                color = 15;
            color = color / 2 + 32768 * (1 - color % 2);
            break;

        case '4':   /* niveaux de gris, chelle linaire (50  100%) */
            color = (intensity - 128) / 8;
            if (color < 0)
                color = 0;
            color = color / 2 + 32768 * (1 - color % 2);
            break;

        case '5':   /* niveaux de gris, chelle logarithmique (0  100%) */
            if (intensity == 0)
                color = 0;
            else {
                color = (int)(2.0 * log((double)(intensity)) / log(2.0));
                color = color / 2 + 32768 * (1 - color % 2);
            }
            break;

        case '8':   /* Affichage en 256 couleurs sur Aurora sans GD2 */
            Aurora_plot(AurMode, (short)x0, (short)y0, (short)Green, (short)Red,
                     (short)Blue);
            return();

        case '9':   /* Affichage en 65536 couleurs sur Q40 sans GD2 */
            Q40_plot(Q40mode, (short)x0, (short)y0, (short)Green, (short)Red,
                     (short)Blue);
            return();
    }

    /* Pour afficher un point on utilise TRAP #3, D0=$2E (16 couleurs et moins)
       ou D0=$5D (256 couleurs et plus avec le GD2): trac d'un rectangle de
       couleur (taille de rectangle=2x1 pixels pour le GD1 ou 1x1 pixels pour
       le GD2) en coordonnes graphiques.
       Ici "qdos3" est utilis en lieu et place de "sd_fill" pour permettre le
       passage du paramtre "color" sous forme d'un mot long (pour affichage 16
       couleurs sur Thor XVI en MODE 12) et l'utilisation du nouveau TRAP du
       GD2.
    */

    if (graphic_mode <= 12)
        inp_par.D1 = color;
    else
        stipple.colour1 = color;

    coord.q_x  = x0;
    coord.q_y  = y0;

    qdos3(&inp_par, &out_par);
}

/* Routine appele en fin d'affichage avant de refermer la fentre graphique */

void QDOS_display_finished(void)
{
    if (Display_Started) {
        beep(5000, 30);

        if (console_in)
            getchar();
    }

    if (opts.DisplayFormat == '9') {
        Q40_cls();
        Q40_mode(1);
    }

    if (opts.DisplayFormat == '8') {
        Aurora_cls();

        /* Remettre l'affichage dans le mode antrieur, */
        /* on suppose que le nombre de couleurs tait de 4. */
        /* Pour la rsolution, on regarde la structure max_size */

        switch (max_size.xsize) {
            case 512:
                if (max_size.ysize == 3/4*512)
                    AurMode = C4R512X384;
                else
                    AurMode = C4R512X256;
                break;
            case 640:
                if (max_size.ysize == 3/4*640)
                    AurMode = C4R640X480;
                else
                    AurMode = C4R640X320;
                break;
            case 768:
                if (max_size.ysize == 3/4*768)
                    AurMode = C4R768X576;
                else
                    AurMode = C4R768X384;
                break;
            case 1024:
                if (max_size.ysize == 3/4*1024)
                    AurMode = C4R1024X768;
                else
                    AurMode = C4R1024X512;
                break;
            default:
                AurMode = C4R512X256;
        }
        Aurora_mode(AurMode);
    }
}

/* Routine de fermeture de la fentre graphique */

void QDOS_display_close(void)
{
    draw_stdout();
}

/*
 * This function accepts a full QDOS/SMS filename (device+path+name) and
 * finds the path and the name of the file. Returns a QDOS error code.
 */

int QDOS_split_filename(char *fname, char *path, char *name)
{
    chanid_t    dumb;
    char        device[40], tmp[80], flag;
    qdirect_t   qlbuf;
    int         i, j, c, offset, ret;

    flag = TRUE;
    j    = 0;

    /*
     * Search for a device name in the form "devN_", "nM_devN_" or "nMM_devN_"
     * (where 0 < N < 9, "dev" is the device name, "n" is the network name, and
     * M/MM is the network address.
     */
    if (strlen(fname) > 8) {            /* First checks for a network name */
        if ((fname[2] == '_' && fname[7] == '_') ||
            (fname[3] == '_' && fname[8] == '_')) {

            if (!isalpha((unsigned char)fname[0]))
                flag = FALSE;

            if (!isdigit((unsigned char)fname[1]))
                flag = FALSE;

            if (fname[3] == '_') {
                j = 4;
                if (!isdigit((unsigned char)fname[2]))
                    flag = FALSE;
            } else
                j = 3;
        }
    }
    if (flag && strlen(fname) > 5) {    /* Now checks for the device name */
            for (i = j; i < j + 3; i++)
                if (!isalpha((unsigned char)fname[i]))
                    flag = FALSE;

            c = (int)(fname[j + 3] - '0');
            if (c < 1 || c > 8 || fname[j + 4] != '_')
                flag = FALSE;

            if (flag) {
                strncpy(device, fname, j + 5);
                device[j + 5] = '\0';
            } else
                return(ERR_BN);
    } else
        return(ERR_BN);

    strcpy(tmp, fname);
    j += 5;

    /*
     * Now search for the path and the file name...
     */
    while (strrchr(tmp, '_') != NULL) {
        offset = strrchr(tmp, '_') - tmp;
        strncpy(name, fname + (offset + 1), strlen(fname) - offset);
        name[strlen(fname) - offset] = '\0';

        if (strlen(fname) - strlen(name) - j > 0) {
            strncpy(path, fname + j, strlen(fname) - strlen(name) - j);
            path[strlen(fname) - strlen(name) - j] = '\0';
        } else {
            path[0] = '\0';
            break;
        }

        strcpy(tmp, device);
        strcat(tmp, path);

        if ((dumb = io_open(tmp, 1)) >= 0) {
            ret = fs_headr(dumb, -1, &qlbuf, 64);
            io_close(dumb);
            if (ret >= 0) {
                c = qlbuf.d_type;
                if (c == QF_DIR_TYPE || c == THOR_DIR_TYPE || c == CST_DIR_TYPE)
                    break;
            }
        }
        tmp[strlen(device) + strlen(path) - 1] = '\0';
    }
    strcpy(tmp, device);
    strcat(tmp, path);
    strcpy(path, tmp);
    return(0);
}

/*
 * Process a default INI file pointed to by the environment variable POVINI.
 * Leaves the global variable Option_String_Ptr set to non-NULL if and only
 * if the environment variable POVINI points to a file which was read.
 */

void QDOS_process_env(void)
{
    if ((Option_String_Ptr = getenv("POVINI")) != NULL) {
        if (!parse_ini_file(Option_String_Ptr)) {
            Warning(0.0, "Could not find '%s' as specified in POVINI environment.\n",
                    Option_String_Ptr);
        Option_String_Ptr = NULL;
        }
    }
    return();
}

/*
 * If QDOS_process_env() processed a default INI file from the POVINI
 * environment variable then do nothing; otherwise search for a default INI
 * file, reading the first one found, from the list:
 * <DATA_USE>_povray_ini
 * POV_LIB_DIR_povray_ini
 */

void QDOS_process_povray_ini(void)
{
    char ini_name[FILE_NAME_MAX];

    if (Option_String_Ptr != NULL)
        return();
    if ((Option_String_Ptr = getenv("DATA_USE")) != NULL) {
        strncpy(ini_name, Option_String_Ptr, FILE_NAME_MAX);
        strncat(ini_name, "povray_ini", FILE_NAME_MAX);
        if (parse_ini_file(ini_name))
            return();
        else
            Option_String_Ptr = NULL;
    }
#ifdef POV_LIB_DIR
    strncpy(ini_name, POV_LIB_DIR, FILE_NAME_MAX);
    strncat(ini_name, "_povray_ini", FILE_NAME_MAX);
    parse_ini_file(ini_name);
#endif
    return();
}

/* QDOS Locate_File */

int Has_Extension(char *name)
{
    char *p;

    if (name != NULL) {
        p = strrchr(name, '.');
        if (p == NULL)
            p = strrchr(name, '_');

        if (p != NULL)
            if ((strlen(name) - (p - name)) <= 4)
                return(TRUE);
    }
    return(FALSE);
}

FILE *Locate_File (char *filename, char *mode, char *ext1, char *ext0, char *buffer, int err_flag)
{
  int i, l1, l2;
  char pathname[FILE_NAME_LENGTH];
  char file0[FILE_NAME_LENGTH];
  char file1[FILE_NAME_LENGTH];
  char file2[FILE_NAME_LENGTH];
  char ext2[FILE_NAME_LENGTH];
  FILE *f;

  if (strlen(ext1) > 1) {
     strcpy(ext2, ext1);
     ext2[0]='_';
  }

  if (Has_Extension(filename))
  {
     l1=l2=0;
  }
  else
  {
     if ((l1 = strlen(ext1)) > 0)
     {
        strcpy(file1, filename);
        strcat(file1, ext1);
     }

     if ((l2 = strlen(ext2)) > 0)
     {
        strcpy(file2, filename);
        strcat(file2, ext2);
     }
  }

  /* Check the current directory first. */
  if (l1)
  {
     if ((f = fopen(file1, mode)) != NULL)
     {
       POV_GET_FULL_PATH(f,file1,buffer);
       return(f);
     }
  }
  if (l2)
  {
     if ((f = fopen(file2, mode)) != NULL)
     {
       POV_GET_FULL_PATH(f,file2,buffer);
       return(f);
     }
  }
  if ((f = fopen(filename, mode)) != NULL)
  {
     POV_GET_FULL_PATH(f,filename,buffer);
     return(f);
  }

  for (i = 0; i < opts.Library_Path_Index; i++)
  {
    strcpy(file0, opts.Library_Paths[i]);
    file0[strlen(file0)+1] = '\0';
    if (file0[strlen(file0) - 1] != DRIVE_SEPARATOR)
      file0[strlen(file0)] = FILENAME_SEPARATOR;

    if (l1)
    {
       strcpy(pathname, file0);
       strcat(pathname, file1);
       if ((f = fopen(pathname, mode)) != NULL)
       {
          POV_GET_FULL_PATH(f,pathname,buffer);
          return(f);
       }
    }

    if (l2)
    {
       strcpy(pathname, file0);
       strcat(pathname, file2);
       if ((f = fopen(pathname, mode)) != NULL)
       {
          POV_GET_FULL_PATH(f,pathname,buffer);
          return(f);
       }
    }
    strcpy(pathname, file0);
    strcat(pathname, filename);
    if ((f = fopen(pathname, mode)) != NULL)
    {
      POV_GET_FULL_PATH(f,pathname,buffer);
      return(f);
    }
  }
 
  if (err_flag)
  {
    if (l1)
    {
      Error_Line("Could not find file '%s%s'\n",filename,ext1);
    }
    else
    {
      Error_Line("Could not find file '%s'\n",filename);
    }
  }
  
  return(NULL);
}

/* Traitement des erreurs de calcul (overflow, sorties des limites...) */

int matherr(struct exception *x)
{
    switch(x->type) {
        case DOMAIN:
        case OVERFLOW:
            x->retval = 1.0e17;
            break;

        case SING:
        case UNDERFLOW:
            x->retval = 0.0;
            break;

        default:
            break;
    }
    return(1);
}
